# -*- coding: utf-8 -*-
#
# Copyright (C) 2019 Esteban J. G. Gabancho.
#
# Flask-SSO-SAML is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.
"""Minimal Flask application example.

SPHINX-START

1. Register your app OneLogin (you might choose any IdP of your liking):

   - ``https://0.0.0.0:5000/saml/acs/onelogin``: SAML Consumer URL
   - ``https://0.0.0.0:5000/saml/sls/onelogin``: SAML Single Logout URL
   - ``https://0.0.0.0:5000/saml/metadata/onelogin``: SAML Audience
   - ``https://0.0.0.0:5000/saml/acs/onelogin``: SAML Recipient

2. Ensure you have ``gunicorn`` package installed:

   .. code-block:: console

      pip install gunicorn

3. Ensure you have ``openssl`` installed in your system (Most of the Linux
   distributions has it by default.).

3. Grab the *Issuer URL*, *SAML 2.0 Endpoint (HTTP)*,
   *SLO Endpoint (HTTP)* and *X.509 Certificate* after registering the
   application (they are under the SSO tab) and add them to your instance
   configuration.

   .. code-block:: console

       $ export IDP_ENTITY_ID='https://app.onelogin.com/saml/metadata/....'
       $ export IDP_SSO_URL='https://myapp-dev.onelogin.com/trust/saml2/http-post/sso/....'
       $ export IDP_SLS_URL='https://myapp-dev.onelogin.com/trust/saml2/http-redirect/slo/....'
       $ export IDP_CERT='one_line_certificate'


4. Install Flask-SSO-SAML and setup the application by running:

   .. code-block:: console

       $ pip install -e .[all]
       $ cd examples
       $ ./app-setup.sh

5. Create the key and the certificate in order to run a HTTPS server:

   .. code-block:: console

       $ openssl genrsa 1024 > instance/ssl.key
       $ openssl req -new -x509 -nodes -sha1 -key instance/ssl.key > instance/ssl.crt

6. Run gunicorn server:

   .. code-block:: console

       $ gunicorn -b :5000 --certfile=./instance/ssl.crt --keyfile=./instance/ssl.key app:app

7. Open in a browser the page `<https://0.0.0.0:5000/>`_.

8. To reset the example application run:

   .. code-block:: console

       $ ./app-teardown.sh

SPHINX-END
"""

from __future__ import absolute_import, print_function

import os

import jinja2
from flask import Flask, redirect, render_template, session

from flask_sso_saml import FlaskSSOSAML


def login_handler(auth, next_url):
    """Reset session things."""
    session['success_slo'] = False
    return next_url


def acs_handler(auth, next_url):
    """Put user info in the session."""
    session['samlUserdata'] = auth.get_attributes()
    session['no_auth_warn'] = not auth.is_authenticated()
    return next_url


def sls_handler(auth, next_url):
    """Set sls in the session."""
    session['success_slo'] = False
    return next_url


def _get_attrs():
    """Get user attributes from the session."""
    if 'samlUserdata' in session:
        paint_logout = True
        if len(session['samlUserdata']) > 0:
            return session['samlUserdata'].items()
    return None


# Create Flask application
app = Flask(__name__)

app.config.update(
    dict(
        SSO_SAML_IDPS={
            'onelogin': {
                'settings': {
                    'idp': {
                        'entityId': os.environ.get('IDP_ENTITY_ID'),
                        'singleSignOnService': {
                            'url': os.environ.get('IDP_SSO_URL')
                        },
                        'singleLogoutService': {
                            'url': os.environ.get('IDP_SLS_URL')
                        },
                        'x509cert': os.environ.get('IDP_CERT'),
                    },
                },
                'login_handler': login_handler,
                'acs_handler': acs_handler,
                'sls_handler': sls_handler,
            }
        },
        SERVER_NAME='0.0.0.0:5000',
        SECRET_KEY='EXAMPLE_APP',
    ))

# Set jinja loader to first grab templates from the app's folder.
app.jinja_loader = jinja2.ChoiceLoader([
    jinja2.FileSystemLoader(
        os.path.join(os.path.dirname(__file__), "templates")), app.jinja_loader
])

FlaskSSOSAML(app)


@app.route('/')
def index():
    """Homepage."""
    attributes = _get_attrs()

    return render_template(
        'index.html',
        attributes=attributes,
        paint_logout=True if attributes else False,
        not_auth_warn=session.get('no_auth_warn'),
        success_slo=session.get('success_slo'),
    )


@app.route('/attrs')
def attrs():
    """User attributes."""
    attributes = _get_attrs()

    return render_template(
        'attrs.html',
        paint_logout=True if attributes else False,
        attributes=attributes)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019 Esteban J. G. Gabancho.
#
# Flask-SSO-SAML is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.
"""Module tests."""

from __future__ import absolute_import, print_function

import os

import pkg_resources
import pytest
from flask import Flask
from mock import patch

from flask_sso_saml import FlaskSSOSAML
from flask_sso_saml.proxies import current_sso_saml


def test_version():
    """Test version import."""
    from flask_sso_saml import __version__
    assert __version__


def test_init():
    """Test extension initialization."""
    app = Flask('testapp')
    ext = FlaskSSOSAML(app)
    assert 'flask-sso-saml' in app.extensions

    app = Flask('testapp')
    ext = FlaskSSOSAML()
    assert 'flask-sso-saml' not in app.extensions
    ext.init_app(app)
    assert 'flask-sso-saml' in app.extensions

    state = app.extensions['flask-sso-saml']
    assert state.prepare_flask_request
    assert callable(state.prepare_flask_request)


def test_app_config(appctx):
    """Test app settings builder."""
    settings_idp1 = current_sso_saml.get_settings('test-idp')

    assert settings_idp1['strict']
    assert settings_idp1['idp']['entityId'] == 'https://test-idp.com'

    # Handlers
    assert callable(current_sso_saml.get_handler('test-idp', 'login_handler'))
    assert callable(current_sso_saml.get_handler('test-idp', 'acs_handler'))
    assert current_sso_saml.get_handler('test-idp', 'settings_handler') is None

    settings_idp2 = current_sso_saml.get_settings('idp-file')

    assert settings_idp2['idp']['entityId'] == "https://login.idp.com"
    assert settings_idp2['sp']['x509cert'] == 'crt\n'
    assert settings_idp2['sp']['privateKey'] == 'key\n'

    reponse_file = pkg_resources.resource_filename(
        __name__, os.path.join('data', 'idp.xml'))
    with open(reponse_file, 'r') as f:
        response = f.read()

    with patch('onelogin.saml2.idp_metadata_parser.urllib2.urlopen'
               ) as urlopen_mock:
        urlopen_mock.return_value = type(
            'Response', (), {
                'read': lambda *args, **kwargs: response
            })()
        settings_idp2 = current_sso_saml.get_settings('idp-url')
        assert settings_idp2['idp']['entityId'] == "https://login.idp.com"


def test_auth(appctx, metadata_response):
    """Test Auth class."""
    with appctx.test_request_context(), pytest.raises(KeyError):
        auth = current_sso_saml.get_auth('wrong-idp')

    with appctx.test_request_context():
        auth = current_sso_saml.get_auth('test-idp')
        assert auth
        assert auth.idp == 'test-idp'



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019 Esteban J. G. Gabancho.
#
# Flask-SSO-SAML is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.
"""Views tests."""

import pytest
from flask import url_for
from mock import patch


def test_wrong_idp(appctx):
    """Test wrong Identity Provider name."""
    login_url = url_for('sso_saml.sso', idp='wrong-idp')
    with appctx.test_client() as client:
        res = client.get(login_url)
        assert res.status_code == 404
        assert b'Identity Provider not found' in res.data


def test_login_error(appctx):
    """Test login error."""
    login_url = url_for('sso_saml.sso', idp='test-idp')
    with appctx.test_client() as client, patch(
            'flask_sso_saml.utils.SAMLAuth.login') as mock_login:
        mock_login.return_value = None
        res = client.get(login_url)
        assert res.status_code == 401


def test_login(appctx):
    """Test SSO requests."""
    login_url = url_for('sso_saml.sso', idp='test-idp')
    with appctx.test_client() as client:
        res = client.get(login_url)
        assert res.status_code == 302
        assert res.location.startswith('https://test-ipd.com/sso?')

    login_url = url_for('sso_saml.sso', idp='test-idp', next='/next_url')
    with appctx.test_client() as client:
        res = client.get(login_url)
        assert res.status_code == 302
        assert res.location.startswith('https://test-ipd.com/sso?')
        assert 'RelayState=%2Fnext_url' in res.location


@pytest.mark.freeze_time('2019-04-19T13:35:47Z')
def test_acs(appctx, sso_response):
    """Test ACS request."""
    acs_url = url_for('sso_saml.acs', idp='test-idp')

    with appctx.test_client() as client:
        res = client.post(acs_url, data=dict(SAMLResponse=''))
        assert res.status_code == 400

    with appctx.test_client() as client:
        res = client.post(acs_url, data=dict(SAMLResponse=sso_response))
        assert res.status_code == 401

    with appctx.test_client() as client, patch(
            'onelogin.saml2.auth.OneLogin_Saml2_Response.is_valid'
    ) as mock_is_valid, patch('flask_sso_saml.utils.SAMLAuth.is_authenticated'
                              ) as mock_is_authenticated:
        mock_is_valid.return_value = True
        mock_is_authenticated.return_value = False
        res = client.post(acs_url, data=dict(SAMLResponse=sso_response))
        assert res.status_code == 403

    with appctx.test_client() as client, patch(
            'onelogin.saml2.auth.OneLogin_Saml2_Response.is_valid'
    ) as mock_is_valid:
        mock_is_valid.return_value = True
        res = client.post(acs_url, data=dict(SAMLResponse=sso_response))
        assert res.status_code == 302


def test_logout(appctx):
    """Test SLO requests."""
    logout_url = url_for('sso_saml.slo', idp='test-idp')
    with appctx.test_client() as client:
        with client.session_transaction() as sess:
            sess['SSO::SAML::NameId'] = 'ID'
            sess['SSO::SAML::SessionIndex'] = 'INDEX'
        res = client.get(logout_url)
        assert res.status_code == 302
        assert res.location.startswith('https://test-ipd.com/slo?')


def test_logout_error(appctx):
    """Test logout error."""
    login_url = url_for('sso_saml.slo', idp='test-idp')
    with appctx.test_client() as client, patch(
            'flask_sso_saml.utils.SAMLAuth.logout') as mock_logout:
        mock_logout.return_value = None
        res = client.get(login_url)
        assert res.status_code == 401


def test_sls(appctx, slo_query_string):
    """Test SLS request."""
    sls_url = url_for('sso_saml.sls', idp='test-idp')
    with appctx.test_client() as client:
        res = client.get(sls_url, query_string=slo_query_string)
        assert res.status_code == 302

    with patch('flask_sso_saml.utils.SAMLAuth.get_errors'
               ) as mock_get_erros, patch(
                   'flask_sso_saml.utils.SAMLAuth.get_last_error_reason'
               ) as mock_get_reason:
        mock_get_erros.return_value = ['bad error']
        mock_get_reason.return_value = 'Test reason'
        with appctx.test_client() as client:
            res = client.get(sls_url, query_string=slo_query_string)
            assert res.status_code == 401
            assert res.json == ['bad error', 'Test reason']


@pytest.mark.freeze_time('2019-04-18')
def test_metadata(appctx, metadata_response):
    """Test metadata request."""
    metadata_url = url_for('sso_saml.metadata', idp='test-idp')
    with appctx.test_client() as client:
        res = client.get(metadata_url)
        assert res.status_code == 200
        assert res.data == metadata_response

    with patch(
            'onelogin.saml2.settings.OneLogin_Saml2_Settings.validate_metadata'
    ) as mock_validate_metadata, patch(
            'flask_sso_saml.utils.SAMLAuth.get_last_error_reason'
    ) as mock_get_reason:
        mock_validate_metadata.return_value = ['bad error', 'worst error']
        mock_get_reason.return_value = 'Test reason'
        with appctx.test_client() as client:
            res = client.get(metadata_url)
            assert res.status_code == 401
            assert res.json == ['bad error', 'worst error', 'Test reason']



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019 Esteban J. G. Gabancho.
#
# Flask-SSO-SAML is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Pytest configuration."""

from __future__ import absolute_import, print_function

import base64
import os
import sys

import pkg_resources
import pytest
from flask import Flask
from onelogin.saml2.utils import OneLogin_Saml2_Utils as saml_utils

from flask_sso_saml import FlaskSSOSAML

PY3 = sys.version_info[0] == 3

if PY3:
    from urllib.parse import urlencode
else:
    from urllib import urlencode


@pytest.fixture(scope='module')
def app():
    """Flask application fixture with FLask-SSO-SAML initialized."""
    app = Flask('testapp')
    app.config.update(
        TESTING=True,
        SECRET_KEY='test-secret-key',
        SERVER_NAME='localhost',
        SSO_SAML_DEFAUTL_LOGIN_HANDLER=lambda auth, next_url: next_url,
        SSO_SAML_IDPS={
            'test-idp': {
                'settings': {
                    'idp': {
                        'entityId': 'https://test-idp.com',
                        'singleSignOnService': {
                            'url': 'https://test-ipd.com/sso',
                        },
                        'singleLogoutService': {
                            'url': 'https://test-ipd.com/slo',
                        },
                        'x509cert': 'cert'
                    }
                },
                'acs_handler': lambda auth, next_url: next_url,
            },
            'idp-file': {
                'settings_file_path':
                pkg_resources.resource_filename(
                    __name__, os.path.join('data', 'idp.xml')),
                'sp_cert_file':
                pkg_resources.resource_filename(
                    __name__, os.path.join('data', 'cert.crt')),
                'sp_key_file':
                pkg_resources.resource_filename(
                    __name__, os.path.join('data', 'cert.key')),
            },
            'idp-url': {
                'settings_url': 'https://test-idp.com/settings'
            },
        })
    FlaskSSOSAML(app)
    return app


@pytest.fixture(scope='module')
def appctx(app):
    """Application context for the current application."""
    with app.app_context():
        yield app


@pytest.fixture(scope='module')
def sso_response():
    """Mock SSO response from Identity Provider."""
    xml_file = pkg_resources.resource_filename(
        __name__, os.path.join('data', 'sso_response.xml'))
    with open(xml_file, 'rb') as f:
        return base64.b64encode(f.read())


@pytest.fixture(scope='module')
def slo_query_string():
    """Mock SLO response from Identity Provider."""
    xml_file = pkg_resources.resource_filename(
        __name__, os.path.join('data', 'slo_response.xml'))
    with open(xml_file) as f:
        slo_response = saml_utils.deflate_and_base64_encode(f.read())
    return urlencode(dict(SAMLResponse=slo_response))


@pytest.fixture(scope='module')
def metadata_response():
    """Metadata response."""
    xml_file = pkg_resources.resource_filename(
        __name__, os.path.join('data', 'metadata.xml'))
    with open(xml_file, 'rb') as f:
        return f.read()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019 Esteban J. G. Gabancho.
#
# Flask-SSO-SAML is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Proxy objects for easier access to application objects."""

from flask import current_app
from werkzeug.local import LocalProxy


def _get_current_sso_saml():
    """Return current state of the SSO SAML extension."""
    return current_app.extensions['flask-sso-saml']


current_sso_saml = LocalProxy(_get_current_sso_saml)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019 Esteban J. G. Gabancho.
#
# Flask-SSO-SAML is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test utils."""

import pytest
from flask import request
from werkzeug.datastructures import MultiDict

from flask_sso_saml.utils import prepare_flask_request


@pytest.mark.parametrize("test_request_ctx,expected", [
    ({}, {
        'get_data': MultiDict([]),
        'http_host': 'localhost',
        'https': 'off',
        'post_data': MultiDict([]),
        'script_name': '/',
        'server_port': None,
    }),
    ({
        'base_url': 'https://tests.com:5000',
        'path': '/bar'
    }, {
        'get_data': MultiDict([]),
        'http_host': 'tests.com:5000',
        'https': 'on',
        'post_data': MultiDict([]),
        'script_name': '/bar',
        'server_port': 5000,
    }),
])
def test_prepare_flask_request(app, test_request_ctx, expected):
    """Test prepare flask request."""
    with app.test_request_context(**test_request_ctx):
        res = prepare_flask_request(request)
        assert res == expected



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019 Esteban J. G. Gabancho.
#
# Flask-SSO-SAML is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Flask extension that provides SSO SAML integration."""

from __future__ import absolute_import, print_function

from .ext import FlaskSSOSAML
from .version import __version__

__all__ = ('__version__', 'FlaskSSOSAML')



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019 Esteban J. G. Gabancho.
#
# Flask-SSO-SAML is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.


"""Python 2/3 compatibility layer."""

from __future__ import absolute_import, print_function

import sys

PY3 = sys.version_info[0] == 3

if PY3:
    string_types = str
    from urllib.parse import urlparse
else:
    string_types = basestring
    from urlparse import urlparse



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019 Esteban J. G. Gabancho.
#
# Flask-SSO-SAML is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Flask extension that provides SSO SAML integration."""

from __future__ import absolute_import, print_function

import collections
import copy
from functools import wraps

from flask import url_for
from onelogin.saml2.idp_metadata_parser import OneLogin_Saml2_IdPMetadataParser
from werkzeug.utils import cached_property, import_string

from . import config
from ._compat import string_types
from .utils import SAMLAuth, prepare_flask_request
from .views import create_blueprint


def _default_config(idp):
    """Default IdP configuration."""
    return dict(
        settings={
            'strict': True,
            'debug': True,
            'sp': {
                'entityId':
                url_for('sso_saml.metadata', idp=idp, _external=True),
                'assertionConsumerService': {
                    'url': url_for('sso_saml.acs', idp=idp, _external=True),
                    'binding':
                    'urn:oasis:names:tc:SAML:2.0:bindings:HTTP-POST',
                },
                'singleLogoutService': {
                    'url':
                    url_for('sso_saml.sls', idp=idp, _external=True),
                    'binding':
                    'urn:oasis:names:tc:SAML:2.0:bindings:HTTP-Redirect',
                },
                'NameIDFormat':
                'urn:oasis:names:tc:SAML:2.0:bindings:HTTP-Redirect',
                'x509cert': '',
                'privateKey': '',
            },
            'idp': {
                'entityId': None,
                'singleSignOnService': {
                    'url': None,
                    'binding':
                    'urn:oasis:names:tc:SAML:2.0:bindings:HTTP-Redirect'
                },
                'singleLogoutService': {
                    'url': None,
                    'binding':
                    'urn:oasis:names:tc:SAML:2.0:bindings:HTTP-Redirect'
                },
                'x509cert': None,
            },
            'security': {
                'authnRequestsSigned': False,
                'failOnAuthnContextMismatch': False,
                'logoutRequestSigned': False,
                'logoutResponseSigned': False,
                'metadataCacheDuration': None,
                'metadataValidUntil': None,
                'nameIdEncrypted': False,
                'requestedAuthnContext': True,
                'requestedAuthnContextComparison': 'exact',
                'signMetadata': False,
                'signatureAlgorithm':
                'http://www.w3.org/2001/04/xmldsig-more#rsa-sha256',
                'wantAssertionsEncrypted': False,
                'wantAssertionsSigned': False,
                'wantMessagesSigned': False,
                'wantNameId': True,
                'wantNameIdEncrypted': False,
                'digestAlgorithm': 'http://www.w3.org/2001/04/xmlenc#sha256'
            },
        },
        settings_file_path=None,
        settings_url=None,
        sp_cert_file=None,
        sp_key_file=None,
        settings_handler=None,
        login_handler=None,
        acs_handler=None,
        logout_handler=None,
        sls_handler=None,
    )


def _cached_configuration(f):
    """Cache the IdP configuration for future use."""
    @wraps(f)
    def inner(self, idp, *args, **kwargs):
        if idp not in self._saml_config:
            self._saml_config[idp] = self._build_configuration(idp)
        return f(self, idp, *args, **kwargs)

    return inner


class _FlaskSSOSAMLState(object):
    """Flask SSO SAML state object."""

    def __init__(self, app):
        """Initialize state."""
        self.app = app
        self._saml_config = {}

    @property
    def url_prefix(self):
        """URL prefix from config."""
        return self.app.config['SSO_SAML_DEFAULT_BLUEPRINT_PREFIX']

    @property
    def metadata_url(self):
        """SSO metadata URL from config."""
        return self.app.config['SSO_SAML_DEFAULT_METADATA_ROUTE']

    @property
    def sso_url(self):
        """SSO SSO URL from config."""
        return self.app.config['SSO_SAML_DEFAULT_SSO_ROUTE']

    @property
    def acs_url(self):
        """SSO ACS URL from config."""
        return self.app.config['SSO_SAML_DEFAULT_ACS_ROUTE']

    @property
    def slo_url(self):
        """SSO SLO URL from config."""
        return self.app.config['SSO_SAML_DEFAULT_SLO_ROUTE']

    @property
    def sls_url(self):
        """SSO SLS URL from config."""
        return self.app.config['SSO_SAML_DEFAULT_SLS_ROUTE']

    @cached_property
    def prepare_flask_request(self):
        """Function to prepare flask request for OneLogin."""
        prep_func = self.app.config['SSO_SAML_PREPARE_FLASK_REQUEST_FUNCTION']
        if isinstance(prep_func, string_types):
            prep_func = import_string(prep_func)
        return prep_func

    @_cached_configuration
    def get_settings(self, idp):
        """Find settings for a particular Identity Provider."""
        return self._saml_config[idp]['settings']

    @_cached_configuration
    def get_handler(self, idp, handler):
        """Get handler for idp."""
        return self._saml_config[idp][handler]

    def get_auth(self, idp):
        """Instantiate the IdP."""
        return SAMLAuth(idp, self.get_settings(idp))

    def _build_configuration(self, idp):
        """Update default config with the ones read from configuration."""
        def update(d, u):
            for k, v in u.items():
                if isinstance(v, collections.Mapping):
                    d[k] = update(d.get(k, {}), v)
                else:
                    d[k] = v
            return d

        def make_handler(handler, default=None):
            handler = handler if handler else default
            return import_string(
                handler
            ) if handler and isinstance(handler, string_types) else handler

        config = _default_config(idp)
        update(config, self.app.config['SSO_SAML_IDPS'][idp])

        # Read IdP config from file or URL if any
        if config['settings_url']:
            external_conf = OneLogin_Saml2_IdPMetadataParser.parse_remote(
                config['settings_url']
            )
            config['settings']['idp'].update(external_conf.get('idp'))

        if config['settings_file_path']:
            with open(config['settings_file_path'], 'r') as idp:
                external_conf = OneLogin_Saml2_IdPMetadataParser.parse(
                    idp.read()
                )
            config['settings']['idp'].update(external_conf.get('idp'))

        # Load certificate and key
        if config['sp_cert_file']:
            with open(config['sp_cert_file'], 'r') as cf:
                cert = cf.read()
            config['settings']['sp']['x509cert'] = cert

        if config['sp_key_file']:
            with open(config['sp_key_file'], 'r') as cf:
                cert = cf.read()
            config['settings']['sp']['privateKey'] = cert

        # Import handlers is present
        config['settings_handler'] = make_handler(
            config['settings_handler'],
            self.app.config.get('SSO_SAML_DEFAUTL_SETTINGS_HANDLER')
        )
        config['login_handler'] = make_handler(
            config['login_handler'],
            self.app.config.get('SSO_SAML_DEFAUTL_LOGIN_HANDLER')
        )
        config['logout_handler'] = make_handler(
            config['logout_handler'],
            self.app.config.get('SSO_SAML_DEFAUTL_LOGOUT_HANDLER')
        )
        config['acs_handler'] = make_handler(
            config['acs_handler'],
            self.app.config.get('SSO_SAML_DEFAUTL_ACS_HANDLER')
        )
        config['sls_handler'] = make_handler(
            config['sls_handler'],
            self.app.config.get('SSO_SAML_DEFAUTL_SLS_HANDLER')
        )

        return config


class FlaskSSOSAML(object):
    """Flask-SSO-SAML extension."""

    def __init__(self, app=None):
        """Extension initialization."""
        if app:
            self.init_app(app)

    def init_app(self, app):
        """Flask application initialization."""
        self.init_config(app)

        state = _FlaskSSOSAMLState(app)

        # Register blueprint and routes
        app.register_blueprint(create_blueprint(state, __name__))

        app.extensions['flask-sso-saml'] = state
        return state

    def init_config(self, app):
        """Initialize configuration."""
        # Only load the configuration, we'll build it lazily because we can
        for k in dir(config):
            if k.startswith('SSO_SAML_'):
                app.config.setdefault(k, getattr(config, k))

        # Set default when needed
        app.config.setdefault(
            'SSO_SAML_PREPARE_FLASK_REQUEST_FUNCTION', prepare_flask_request
        )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019 Esteban J. G. Gabancho.
#
# Flask-SSO-SAML is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.
"""SAML integration functions."""

from functools import wraps

from flask import Blueprint, abort, current_app, jsonify, make_response, \
    redirect, request, session
from werkzeug.local import LocalProxy

from .proxies import current_sso_saml


def verify_idp(f):
    """Check if the Identity Provider is correctly exists."""
    @wraps(f)
    def inner(idp, *args, **kwargs):
        try:
            return f(
                idp=idp, auth=current_sso_saml.get_auth(idp), *args, **kwargs)
        except KeyError as e:
            # IdP name not found inside the configuration
            return abort(404, 'Identity Provider not found')

    return inner


@verify_idp
def metadata(idp, auth):
    """Expose XML configuration of the Service Provider (us)."""
    current_app.logger.debug('Handling metadata request for {}'.format(idp))

    settings = auth.get_settings()
    sp_metadata = settings.get_sp_metadata()
    errors = settings.validate_metadata(sp_metadata)

    if errors:
        error_reason = auth.get_last_error_reason()
        if error_reason:
            errors.append(auth.get_last_error_reason())
            current_app.logger.error(
                'Handling metadata request: {}'.format(errors))
        return jsonify(errors), 401
    else:
        current_app.logger.debug(
            'Metadata request response: {}'.format(sp_metadata))
        resp = make_response(sp_metadata, 200)
        resp.headers['Content-Type'] = 'text/xml'
        return resp


@verify_idp
def sso(idp, auth):
    """Send user to IdP login page (SAML single sign-on)."""
    current_app.logger.debug('SSO SAML for {}'.format(idp))

    next_url = request.args.get('next', request.referrer) or request.host_url
    login = auth.login(return_to=next_url)

    current_app.logger.debug('SSO Request XML: \n{}'.format(
        auth.get_last_request_xml()))
    current_app.logger.debug(
        'Redirecting to "{}" to initiate login'.format(login))

    if not login:
        # This should never happen, but just in case
        abort(401)

    return redirect(login)


@verify_idp
def acs(idp, auth):
    """Authorized handler callback (Assertion Consumer Service).

    It gets called by the IdP with SAML assertion when authentication has been
    performed.
    """
    try:
        # TODO https://github.com/onelogin/python3-saml/issues/39 ?
        auth.process_response()
    except Exception:  # TODO better exception handling
        return abort(400)

    errors = auth.get_errors()

    if errors:
        error_reason = auth.get_last_error_reason()
        if error_reason:
            errors.append(auth.get_last_error_reason())
            current_app.logger.error('Handling ACS request: {}'.format(errors))
        return jsonify(errors), 401

    if not auth.is_authenticated():
        return abort(403)

    # Set SSO specific IdP metadata in the session, (used later in slo)
    session[current_app.
            config['SSO_SAML_SESSION_KEY_NAME_ID']] = auth.get_nameid()
    session[current_app.config[
        'SSO_SAML_SESSION_KEY_SESSION_INDEX']] = auth.get_session_index()

    next_url = auth.acs_handler(request.form.get('RelayState')) or '/'

    return redirect(next_url)


@verify_idp
def slo(idp, auth):
    """Send user to IdP logout page (SAML single logout)."""
    name_id = session.get(current_app.config['SSO_SAML_SESSION_KEY_NAME_ID'])
    session_index = session.get(
        current_app.config['SSO_SAML_SESSION_KEY_SESSION_INDEX'])

    next_url = request.args.get('next', request.referrer) or request.host_url

    logout = auth.logout(
        return_to=next_url, name_id=name_id, session_index=session_index)

    current_app.logger.debug('SLO Request XML: \n{}'.format(
        auth.get_last_request_xml()))
    current_app.logger.debug(
        'Redirecting to "{}" to initiate logout'.format(logout))

    if not logout:
        # This should never happen, but just in case
        abort(401)

    return redirect(logout)


@verify_idp
def sls(idp, auth):
    """Logout handler callback (Single Logout Service).

    It Consumes LogoutResponse from IdP when logout has been performed.
    """
    # Process the SLO message received from IdP
    next_url = auth.process_slo(delete_session_cb=lambda: session.clear())

    errors = auth.get_errors()
    if errors:
        error_reason = auth.get_last_error_reason()
        if error_reason:
            errors.append(auth.get_last_error_reason())
            current_app.logger.error('Handling SLS request: {}'.format(errors))
        return jsonify(errors), 401

    next_url = auth.sls_handler(next_url) or '/'

    return redirect(next_url)


def create_blueprint(state, import_name):
    """Create the SSO SAML extension blueprint."""
    bp = Blueprint('sso_saml', import_name, url_prefix=state.url_prefix)

    bp.add_url_rule(
        state.metadata_url, endpoint='metadata', view_func=metadata)

    bp.add_url_rule(
        state.sso_url, methods=['GET', 'POST'], endpoint='sso', view_func=sso)

    bp.add_url_rule(
        state.acs_url, methods=['GET', 'POST'], endpoint='acs', view_func=acs)

    bp.add_url_rule(
        state.slo_url, methods=['GET', 'POST'], endpoint='slo', view_func=slo)

    bp.add_url_rule(
        state.sls_url, endpoint='sls', view_func=sls)

    return bp



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019 Esteban J. G. Gabancho.
#
# Flask-SSO-SAML is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Version information for Flask-SSO-SAML.

This file is imported by ``flask_sso_saml.__init__``,
and parsed by ``setup.py``.
"""

from __future__ import absolute_import, print_function

__version__ = '0.1.0'



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019 Esteban J. G. Gabancho.
#
# Flask-SSO-SAML is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.
"""Flask extension that provides SSO SAML integration."""

SSO_SAML_SESSION_KEY_NAME_ID = 'SSO::SAML::NameId'
"""Key name to store the SSO Name ID in the session."""

SSO_SAML_SESSION_KEY_SESSION_INDEX = 'SSO::SAML::SessionIndex'
"""Key name to store the SSO Session Index in the session."""

SSO_SAML_PREPARE_FLASK_REQUEST_FUNCTION = \
    'flask_sso_saml.utils.prepare_flask_request'
"""Default function to prepare the flask request to be sent to the IdP.

If the server is behind proxys or balancers, this function might need to be
updated to use the HTTP_X_FORWARDED fields.
"""

SSO_SAML_IDPS = {}
"""SSO SAML Identity provider configuration.

This configuration variable can be used to describe the endpoints used for the
different IdPs and their settings.

The structure of the dictionary is as follows:

.. code-block:: python

    SSO_SAML_IDPS = {
        'idp-name': {
            'settings': {
                'strict': True,
                'debug': True,
                'sp': {
                    'entityId':
                    flask.url_for(
                        'sso_saml.metadata', idp='idp-name', _external=True),
                    'assertionConsumerService': {
                        'url':
                        flask.url_for(
                            'sso_saml.acs', idp='idp-name', _external=True),
                        'binding':
                        'urn:oasis:names:tc:SAML:2.0:bindings:HTTP-POST',
                    },
                    'singleLogoutService': {
                        'url':
                        flask.url_for(
                            'sso_saml.sls', idp='idp-name', _external=True),
                        'binding':
                        'urn:oasis:names:tc:SAML:2.0:bindings:HTTP-Redirect',
                    },
                    'attributeConsumingService': {
                        # TODO
                    },
                    'NameIDFormat':
                    'urn:oasis:names:tc:SAML:2.0:bindings:HTTP-Redirect',
                    'x509cert':
                    '',
                    'privateKey':
                    ''
                },
                'idp': {
                    'entityId': '',  #TODO
                    'singleSignOnService': {
                        'url': '',  # TODO
                        'binding':
                        'urn:oasis:names:tc:SAML:2.0:bindings:HTTP-Redirect'
                    },
                    'singleLogoutService': {
                        'url': '',  # TODO
                        'binding':
                        'urn:oasis:names:tc:SAML:2.0:bindings:HTTP-Redirect'
                    },
                    'x509cert': ''  # TODO oneliner
                },
                'security': {
                    'authnRequestsSigned': False,
                    'failOnAuthnContextMismatch': False,
                    'logoutRequestSigned': False,
                    'logoutResponseSigned': False,
                    'metadataCacheDuration': None,
                    'metadataValidUntil': None,
                    'nameIdEncrypted': False,
                    'requestedAuthnContext': True,
                    'requestedAuthnContextComparison': 'exact',
                    'signMetadata': False,
                    'signatureAlgorithm':
                    'http://www.w3.org/2001/04/xmldsig-more#rsa-sha256',
                    'wantAssertionsEncrypted': False,
                    'wantAssertionsSigned': False,
                    'wantMessagesSigned': False,
                    'wantNameId': True,
                    'wantNameIdEncrypted': False,
                    'digestAlgorithm':
                    'http://www.w3.org/2001/04/xmlenc#sha256'
                },
                'contactPerson': {
                    # TODO
                },
                'organization': {
                    # TODO
                },
            },
            'settings_file_path': '/full/path/to/directory',
            'settings_url': 'https://...',
            'settings_handler': '...',
            'login_handler': '...',
            'acs_handler': '...',
            'logout_handler': '...',
            'sls_handler': '...',
        }
    }

:param settings: Setting dictionary compatible with OneLogin. Defatult values
    are show in the example. At least the following need to be provided.
    Under the Identity Provider, ``idp``, settings:
    - ``entityId``
    - ``singleSignOnService.url``
    - ``singleLogoutService.url``
    - ``x509cert``
    If you want to provide X.509cert and privateKey of the Service Provider,
    instead of via file path, you should define the following under ``sp``:
    - ``x509cert``
    - ``privateKey``
    - ``x509certNew`` (only upon certificate update)
    You might also want to provide information about ``contactPerson`` and
    ``organization`` see https://github.com/onelogin/python3-saml#settings for
    more information.

:param settings_file_path: Base folder where ``setting.json`` and
    ``advanced_settings.json`` are located. This parameter will override the
    URL parameter and update the settings found inside the configuration
    variable if any.

:param settings_url: The URL to the IdPs metadata. This parameter will update
    the values found inside the configuration variable if any.

:param settings_handler: Import path to settings handler. Python
    callable which receives two parameters, an instance of ``SAMLAuth`` and the
    current settings returned by``OneLogin_Saml2_Auth.get_settings``.
    Typically returns the settings, although it is not mandatory.

:param login_handler: Import path to login handler. Python callable which
    receives two parameters, an instance of ``SAMLAuth`` and the next url that
    the user will be redirected to. It gets called after
    ``OneLogin_Saml2_Auth.login``.
    Typically returns the next URL too, although it is not mandatory.

:param acs_handler: Import path to ACS handler. Python callable which receives
    two parameters, an instance of ``SAMLAuth`` and the next url that the user
    will be redirected to. It gets called after
    ``OneLogin_Saml2_Auth.process_response``.
    Typically returns the next URL too, although it is not mandatory.

:param logout_handler: Import path to logout handler. Python callable which
    receives two parameters, an instance of ``SAMLAuth`` and the next url that
    the user will be redirected to. It gets called after
    ``OneLogin_Saml2_Auth.logout``.
    Typically returns the next URL too, although it is not mandatory.

:param sls_handler: Import path to SLS handler. Python callable which
    receives two parameters, an instance of ``SAMLAuth`` and the next url that
    the user will be redirected to. It gets called after
    ``OneLogin_Saml2_Auth.process_slo``.
    Typically returns the next URL too, although it is not mandatory.
"""


# Default handlers

SSO_SAML_DEFAUTL_SETTINGS_HANDLER = None
"""Default settings request handler."""

SSO_SAML_DEFAUTL_LOGIN_HANDLER = None
"""Default login request handler."""

SSO_SAML_DEFAUTL_ACS_HANDLER = None
"""Default ACS request handler."""

SSO_SAML_DEFAUTL_LOGOUT_HANDLER = None
"""Default logut request handler."""

SSO_SAML_DEFAUTL_SLS_HANDLER = None
"""Default SLS request handler."""

# Blueprint and routes default configuration

SSO_SAML_DEFAULT_BLUEPRINT_PREFIX = '/saml'
"""Base URL for the extensions endpoint."""

SSO_SAML_DEFAULT_METADATA_ROUTE = '/metadata/<idp>'
"""URL route for the metadata request."""

SSO_SAML_DEFAULT_SSO_ROUTE = '/sso/<idp>'
"""URL route for the SP login."""

SSO_SAML_DEFAULT_ACS_ROUTE = '/acs/<idp>'
"""URL route to handle the IdP login request."""

SSO_SAML_DEFAULT_SLO_ROUTE = '/slo/<idp>'
"""URL route for the SP logout."""

SSO_SAML_DEFAULT_SLS_ROUTE = '/sls/<idp>'
"""URL route to handle the IdP logout request."""



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019 Esteban J. G. Gabancho.
#
# Flask-SSO-SAML is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test example app."""

import os
import signal
import subprocess
import time
from os.path import abspath, dirname, join

import pytest


@pytest.yield_fixture
def example_app():
    """Example app fixture."""
    current_dir = os.getcwd()

    # Go to example directory
    project_dir = dirname(dirname(abspath(__file__)))
    exampleapp_dir = join(project_dir, 'examples')
    os.chdir(exampleapp_dir)

    # Setup application
    assert subprocess.call('./app-setup.sh', shell=True) == 0

    # Start example app
    webapp = subprocess.Popen(
        'FLASK_APP=app.py flask run --debugger -p 5000',
        stdout=subprocess.PIPE, preexec_fn=os.setsid, shell=True)
    time.sleep(10)

    yield webapp

    # Stop server
    os.killpg(webapp.pid, signal.SIGTERM)

    # Tear down example app
    subprocess.call('./app-teardown.sh', shell=True)

    # Return to the original directory
    os.chdir(current_dir)


def test_example_app_role_admin(example_app):
    """Test example app."""
    cmd = 'curl http://0.0.0.0:5000/'
    output = subprocess.check_output(cmd, shell=True)
    assert b'Login' in output



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019 Esteban J. G. Gabancho.
#
# Flask-SSO-SAML is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Flask extension that provides SSO SAML integration."""

import os

from setuptools import find_packages, setup

readme = open('README.rst').read()
history = open('CHANGES.rst').read()

tests_require = [
    'check-manifest>=0.25',
    'coverage>=4.0',
    'isort>=4.3.3',
    'mock>=2.0.0',
    'pydocstyle>=2.0.0',
    'pytest-cov>=2.5.1',
    'pytest-freezegun>=0.2.0',
    'pytest-pep8>=1.0.6',
    'pytest>=3.8.2',
]

extras_require = {
    'docs': [
        'Sphinx>=1.5.1',
    ],
    'tests': tests_require,
}

extras_require['all'] = []
for reqs in extras_require.values():
    extras_require['all'].extend(reqs)

setup_requires = [
    'Babel>=1.3',
    'pytest-runner>=3.0.0,<5',
]

install_requires = [
    'Flask>=1.0',
    'python3-saml>=1.5.0',
]

packages = find_packages()


# Get the version string. Cannot be done with import!
g = {}
with open(os.path.join('flask_sso_saml', 'version.py'), 'rt') as fp:
    exec(fp.read(), g)
    version = g['__version__']

setup(
    name='flask-sso-saml',
    version=version,
    description=__doc__,
    long_description=readme + '\n\n' + history,
    keywords='Flask SSO SAML',
    license='MIT',
    author='Esteban J. G. Gabancho',
    author_email='egabancho@gmail.com',
    url='https://github.com/egabancho/flask-sso-saml',
    packages=packages,
    zip_safe=False,
    include_package_data=True,
    platforms='any',
    extras_require=extras_require,
    install_requires=install_requires,
    setup_requires=setup_requires,
    tests_require=tests_require,
    classifiers=[
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Development Status :: 1 - Planning',
    ],
)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019 Esteban J. G. Gabancho.
#
# Flask-SSO-SAML is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.
"""Utility functions."""

from functools import wraps

from flask import current_app, request
from onelogin.saml2.auth import OneLogin_Saml2_Auth

from ._compat import urlparse
from .proxies import current_sso_saml


def prepare_flask_request(request):
    """Prepare OneLogin-friendly request."""
    # If server is behind proxys or balancers use the HTTP_X_FORWARDED fields
    uri_data = urlparse(request.url)
    return {
        'get_data': request.args.copy(),
        'http_host': request.host,
        'https': 'on' if request.scheme == 'https' else 'off',
        'post_data': request.form.copy(),
        'script_name': request.path,
        'server_port': uri_data.port,
        # Uncomment if using ADFS as IdP,
        # https://github.com/onelogin/python-saml/pull/144
        # 'lowercase_urlencoding': True,
    }


def run_handler(handler_name):
    """Find a handler inside configuration and call it."""
    def decorated(f):
        @wraps(f)
        def inner(self, *args, **kwargs):
            res = f(self, *args, **kwargs)
            handler = current_sso_saml.get_handler(self.idp, handler_name)
            if handler:
                return handler(self, res)
            return res
        return inner
    return decorated


class SAMLAuth(OneLogin_Saml2_Auth):
    """Encapsulate OneLogin SP SAML instance."""

    def __init__(self, idp, settings, *args, **kwargs):
        """Initialization."""
        self.idp = idp
        self._settings = settings
        req = current_sso_saml.prepare_flask_request(request)
        super(SAMLAuth, self).__init__(req, self._settings, *args, **kwargs)

    @run_handler('settings_handler')
    def get_settings(self):
        """Get settings info and call handler.

        :return: ``OneLogin_Saml2_Setting`` object
        """
        settings = super(SAMLAuth, self).get_settings()
        return settings

    @run_handler('login_handler')
    def login(self, *args, **kwargs):
        """Wrapper around ``OneLogin_Saml2_Auth.login``."""
        next_url = super(SAMLAuth, self).login(*args, **kwargs)
        return next_url

    @run_handler('logout_handler')
    def logout(self, *args, **kwargs):
        """Wrapper around ``OneLogin_Saml2_Auth.logout``."""
        next_url = super(SAMLAuth, self).logout(*args, **kwargs)
        return next_url

    @run_handler('acs_handler')
    def acs_handler(self, next_url):
        """Call ACS handler from config."""
        return next_url

    @run_handler('sls_handler')
    def sls_handler(self, next_url):
        """Call SLS handler from config."""
        return next_url
